/*-
# X-BASED RUBIK'S CUBE(tm)
#
#  RubikGL.c
#
###
#
#  Copyright (c) 2003 - 2007	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 *event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * This mode shows an auto-solving rubik's cube "puzzle". If somebody
 * intends to make a game or something based on this code, please let me
 * know first, my e-mail address is provided in this comment. Marcelo.
 *
 * Thanks goes also to Brian Paul for making it possible and inexpensive
 * to use OpenGL at home.
 *
 * Since I'm not a native English speaker, my apologies for any grammatical
 * mistakes.
 *
 * My e-mail address is
 * mfvianna@centroin.com.br
 *
 * Marcelo F. Vianna (Jul-31-1997)
 */

#ifdef HAVE_OPENGL

/* Methods file for RubikGL */

#include "RubikP.h"
#include "RubikGLP.h"

static float front_shininess[] =
{60.0};
static float front_specular[] =
{0.7, 0.7, 0.7, 1.0};
static float ambient[] =
{0.0, 0.0, 0.0, 1.0};
static float diffuse[] =
{1.0, 1.0, 1.0, 1.0};
static float position0[] =
{1.0, 1.0, 1.0, 0.0};
static float position1[] =
{-1.0, -1.0, 1.0, 0.0};
static float lmodel_ambient[] =
{0.5, 0.5, 0.5, 1.0};
static float lmodel_twoside[] =
{GL_TRUE};

static float MaterialMagenta[] =
{0.7, 0.0, 0.7, 1.0};
static float MaterialCyan[] =
{0.0, 0.7, 0.7, 1.0};
static float MaterialGray[] =
{0.2, 0.2, 0.2, 1.0};

/*-
 * Examine facet 0 on each face, its 4 movements (well only 2 since the
 * other 2 will be opposites) and translate it into slice movements).
 * CW = DEEP Depth CCW == SHALLOW Depth with reference to faces 0, 1, and 2
 */
static RubikLoc rotateSlice[MAXFACES][MAXORIENT / 2] =
{
	{
		{1, CCW},
		{2, CW},
	},
	{
		{2, CW},
		{0, CCW},
	},
	{
		{1, CCW},
		{0, CCW},
	},
	{
		{2, CCW},
		{0, CCW},
	},
	{
		{1, CCW},
		{2, CCW},
	},
	{
		{1, CCW},
		{0, CW},
	}
};

static int rotateDir[MAXFACES][MAXORIENT] =
{
	{-1, -3, 1, 3},
	{-3, 2, 3, -2},
	{-1, 2, 1, -2},
	{3, 2, -3, -2},
	{-1, 3, 1, -3},
	{-1, -2, 1, 2}
};

/*-
 * Color labels mapping:
 * =====================
 *
 *             +-----------+
 *             |0-->       |
 *             ||          |
 *             |v  TOP(0)  |
 *             |   RED     |
 *             |          8|
 * +-----------+-----------+-----------+
 * |0-->       |0-->       |0-->       |
 * ||          ||          ||          |
 * |v  LEFT(1) |v FRONT(2) |v RIGHT(3) |
 * |   YELLOW  |  WHITE    |  GREEN    |
 * |          8|          8|          8|
 * +-----------+-----------+-----------+
 *             |0-->       |
 *             ||          |
 *             |v BOTTOM(4)|
 *             |  ORANGE   |
 *             |          8|
 *             +-----------+             +---+---+---+
 *             |0-->       |             | 0 | 1 | 2 |
 *             ||          |             |--xxxxx(N)-+
 *             |v  BACK(5) |             | 3 | 4 | 5 |
 *             |   BLUE    |             +---+---+---+
 *             |          8|             | 6 | 7 | 8 |
 *             +-----------+             +---+---+---+
 *
 *  Map to 3d
 *  FRONT  => X, Y
 *  BACK   => X, Y
 *  LEFT   => Z, Y
 *  RIGHT  => Z, Y
 *  TOP    => X, Z
 *  BOTTOM => X, Z
 */

static Boolean madeCurrent = False;

#ifdef WINVER
static HGLRC hRC = NULL;
#else
static GLXContext *glXContext = (GLXContext *) NULL;
static Boolean SetValuesPuzzleGL(Widget current, Widget request, Widget renew);
static void ResizePuzzleGL(RubikGLWidget w);
static void InitializePuzzleGL(Widget request, Widget renew);
static void ExposePuzzleGL(Widget renew, XEvent *event, Region region);
static void MovePuzzleGLTl(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);
static void MovePuzzleGLTop(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);
static void MovePuzzleGLTr(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);
static void MovePuzzleGLLeft(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);
static void MovePuzzleGLRight(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);
static void MovePuzzleGLBl(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);
static void MovePuzzleGLBottom(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);
static void MovePuzzleGLBr(RubikGLWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsPuzzleGL[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>0x2E: Speed()\n\
 <KeyPress>0x3E: Speed()\n\
 <KeyPress>0x3C: Slow()\n\
 <KeyPress>0x2C: Slow()\n\
 Shift<KeyPress>2: Sound()\n\
 <KeyPress>KP_Divide: MoveCcw()\n\
 <KeyPress>R5: MoveCcw()\n\
 <KeyPress>Home: MoveTl()\n\
 <KeyPress>KP_7: MoveTl()\n\
 <KeyPress>R7: MoveTl()\n\
 <KeyPress>Up: MoveTop()\n\
 <KeyPress>osfUp: MoveTop()\n\
 <KeyPress>KP_Up: MoveTop()\n\
 <KeyPress>KP_8: MoveTop()\n\
 <KeyPress>R8: MoveTop()\n\
 <KeyPress>Prior: MoveTr()\n\
 <KeyPress>KP_9: MoveTr()\n\
 <KeyPress>R9: MoveTr()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <KeyPress>Begin: MoveCw()\n\
 <KeyPress>KP_5: MoveCw()\n\
 <KeyPress>R11: MoveCw()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <KeyPress>End: MoveBl()\n\
 <KeyPress>KP_1: MoveBl()\n\
 <KeyPress>R13: MoveBl()\n\
 <KeyPress>Down: MoveBottom()\n\
 <KeyPress>osfDown: MoveBottom()\n\
 <KeyPress>KP_Down: MoveBottom()\n\
 <KeyPress>KP_2: MoveBottom()\n\
 <KeyPress>R14: MoveBottom()\n\
 <KeyPress>Next: MoveBr()\n\
 <KeyPress>KP_3: MoveBr()\n\
 <KeyPress>R15: MoveBr()\n\
 <Btn1Down>: Select()\n\
 <Btn1Up>: Release()\n\
 <Btn2Down>: PracticeMaybe()\n\
 <Btn2Down>(2+): Practice2()\n\
 <Btn3Down>: RandomizeMaybe()\n\
 <Btn3Down>(2+): Randomize2()\n\
 <KeyPress>g: Get()\n\
 <KeyPress>w: Write()\n\
 <KeyPress>u: Undo()\n\
 <KeyPress>r: Redo()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>z: Randomize()\n\
 <KeyPress>s: Solve()\n\
 <KeyPress>p: Practice()\n\
 <KeyPress>i: Increment()\n\
 <KeyPress>d: Decrement()\n\
 <KeyPress>x: IncrementX()\n\
 <KeyPress>y: IncrementY()\n\
 <KeyPress>f: IncrementZ()\n\
 <KeyPress>o: Orientize()\n\
 <KeyPress>v: View()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListPuzzleGL[] =
{
	{(char *) "Quit", (XtActionProc) QuitPuzzle},
	{(char *) "Hide", (XtActionProc) HidePuzzle},
	{(char *) "MoveCcw", (XtActionProc) MovePuzzleCcw},
	{(char *) "MoveTl", (XtActionProc) MovePuzzleGLTl},
	{(char *) "MoveTop", (XtActionProc) MovePuzzleGLTop},
	{(char *) "MoveTr", (XtActionProc) MovePuzzleGLTr},
	{(char *) "MoveLeft", (XtActionProc) MovePuzzleGLLeft},
	{(char *) "MoveCw", (XtActionProc) MovePuzzleCw},
	{(char *) "MoveRight", (XtActionProc) MovePuzzleGLRight},
	{(char *) "MoveBl", (XtActionProc) MovePuzzleGLBl},
	{(char *) "MoveBottom", (XtActionProc) MovePuzzleGLBottom},
	{(char *) "MoveBr", (XtActionProc) MovePuzzleGLBr},
	{(char *) "Select", (XtActionProc) SelectPuzzle},
	{(char *) "Release", (XtActionProc) ReleasePuzzle},
	{(char *) "PracticeMaybe", (XtActionProc) PracticePuzzleMaybe},
	{(char *) "Practice2", (XtActionProc) PracticePuzzle2},
	{(char *) "RandomizeMaybe", (XtActionProc) RandomizePuzzleMaybe},
	{(char *) "Randomize2", (XtActionProc) RandomizePuzzle2},
	{(char *) "Get", (XtActionProc) GetPuzzle},
	{(char *) "Write", (XtActionProc) WritePuzzle},
	{(char *) "Undo", (XtActionProc) UndoPuzzle},
	{(char *) "Redo", (XtActionProc) RedoPuzzle},
	{(char *) "Clear", (XtActionProc) ClearPuzzle},
	{(char *) "Randomize", (XtActionProc) RandomizePuzzle},
	{(char *) "Solve", (XtActionProc) SolvePuzzle},
	{(char *) "Practice", (XtActionProc) PracticePuzzle},
	{(char *) "Orientize", (XtActionProc) OrientizePuzzle},
	{(char *) "Decrement", (XtActionProc) DecrementPuzzle},
	{(char *) "Increment", (XtActionProc) IncrementPuzzle},
	{(char *) "IncrementX", (XtActionProc) IncrementXPuzzle},
	{(char *) "IncrementY", (XtActionProc) IncrementYPuzzle},
	{(char *) "IncrementZ", (XtActionProc) IncrementZPuzzle},
	{(char *) "View", (XtActionProc) ViewPuzzle},
	{(char *) "Speed", (XtActionProc) SpeedPuzzle},
	{(char *) "Slow", (XtActionProc) SlowPuzzle},
	{(char *) "Sound", (XtActionProc) SoundPuzzle},
	{(char *) "Enter", (XtActionProc) EnterPuzzle},
	{(char *) "Leave", (XtActionProc) LeavePuzzle}
};

static XtResource resourcesPuzzleGL[] =
{
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(RubikWidget, core.width),
	 XtRString, (caddr_t) "250"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(RubikWidget, core.height),
	 XtRString, (caddr_t) "400"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(RubikWidget, rubik.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(RubikWidget, rubik.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNframeColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(RubikWidget, rubik.frameColor),
	 XtRString, (caddr_t) "cyan" /*XtDefaultForeground*/},
	{XtNfaceColor0, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[0]),
	 XtRString, (caddr_t) "Red"},
	{XtNfaceColor1, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[1]),
	 XtRString, (caddr_t) "Yellow"},
	{XtNfaceColor2, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[2]),
	 XtRString, (caddr_t) "White"},
	{XtNfaceColor3, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[3]),
	 XtRString, (caddr_t) "Green"},
	{XtNfaceColor4, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[4]),
	 XtRString, (caddr_t) "Orange"},
	{XtNfaceColor5, XtCLabel, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.faceName[5]),
	 XtRString, (caddr_t) "Blue"},
	{XtNpieceBorder, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(RubikWidget, rubik.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultForeground*/},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.delay),
	 XtRString, (caddr_t) "10"},
	{XtNsound, XtCSound, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.sound),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmoveSound, XtCMoveSound, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.moveSound),
         XtRString, (caddr_t) MOVESOUND},
	{XtNfont, XtCFont, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.font),
	 XtRString, (caddr_t) "9x15bold"},
	{XtNview, XtCView, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.view),
	 XtRString, (caddr_t) "1"},
	{XtNsizex, XtCSizeX, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.sizex),
	 XtRString, (caddr_t) "3"},	/* DEFAULTFACETS */
	{XtNsizey, XtCSizeY, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.sizey),
	 XtRString, (caddr_t) "3"},	/* DEFAULTFACETS */
	{XtNsizez, XtCSizeZ, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.sizez),
	 XtRString, (caddr_t) "3"},	/* DEFAULTFACETS */
	{XtNorient, XtCOrient, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.orient),
	 XtRString, (caddr_t) "FALSE"},	/* DEFAULTORIENT */
	{XtNpractice, XtCPractice, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.practice),
	 XtRString, (caddr_t) "TRUE"}, /* DEFAULTPRACTICE */
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.userName),
	 XtRString, (caddr_t) ""},
	{XtNscoreFile, XtCScoreFile, XtRString, sizeof (String),
	 XtOffset(RubikWidget, rubik.scoreFile),
	 XtRString, (caddr_t) ""},
	{XtNscoreOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.scoreOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNversionOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.versionOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.menu),
	 XtRString, (caddr_t) "999"}, /* ACTION_IGNORE */
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNcheat, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.cheat),
	 XtRString, (caddr_t) "FALSE"},
	{XtNface, XtCFace, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.currentFace),
	 XtRString, (caddr_t) "-1"},
	{XtNpos, XtCPos, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.currentPosition),
	 XtRString, (caddr_t) "-1"},
	{XtNdirection, XtCDirection, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.currentDirection),
	 XtRString, (caddr_t) "-1"},
	{XtNcontrol, XtCControl, XtRBoolean, sizeof (Boolean),
	 XtOffset(RubikWidget, rubik.currentControl),
	 XtRString, (caddr_t) "FALSE"},
	{XtNfast, XtCFast, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.currentFast),
	 XtRString, (caddr_t) "1"},
	{XtNpixmapSize, XtCPixmapSize, XtRInt, sizeof (int),
	 XtOffset(RubikWidget, rubik.pixmapSize),
	 XtRString, (caddr_t) "64"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(RubikWidget, rubik.select),
	 XtRCallback, (caddr_t) NULL}
};

RubikGLClassRec rubikGLClassRec =
{
	{
		(WidgetClass) & rubikClassRec,	/* superclass */
		(char *) "RubikGL",	/* class name */
		sizeof (RubikGLRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializePuzzleGL,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListPuzzleGL,	/* actions */
		XtNumber(actionsListPuzzleGL),	/* num actions */
		resourcesPuzzleGL,	/* resources */
		XtNumber(resourcesPuzzleGL),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		NULL,		/* destroy */
		(XtWidgetProc) ResizePuzzleGL,	/* resize */
		(XtExposeProc) ExposePuzzleGL,	/* expose */
		(XtSetValuesFunc) SetValuesPuzzleGL,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		XtInheritAcceptFocus,	/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsPuzzleGL,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	},
	{
		0		/* ignore */
	}
};

WidgetClass rubikGLWidgetClass = (WidgetClass) & rubikGLClassRec;
#endif

#ifdef DEBUG
void
printCube(RubikGLWidget w)
{
	int face, position, sizeOfRow, sizeOfColumn;

	for (face = 0; face < MAXFACES; face++) {
		faceSizes(w, face, &sizeOfRow, &sizeOfColumn);
		for (position = 0; position < sizeOfRow * sizeOfColumn; position++) {
			(void) printf("%d %d  ",
				w->rubik.cubeLoc[face][position].face,
				w->rubik.cubeLoc[face][position].rotation);
			if (!((position + 1) % sizeOfRow))
				(void) printf("\n");
		}
		(void) printf("\n");
	}
	(void) printf("\n");
}

#endif

static void
pickColor(RubikGLWidget w, int C, int mono)
{
	float Material[] = {1.0, 1.0, 1.0, 1.0};

	switch (C) {
	case TOP_FACE:
	case LEFT_FACE:
	case FRONT_FACE:
	case RIGHT_FACE:
	case BOTTOM_FACE:
	case BACK_FACE:
		break;
	case TEST_FACE:
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,
			MaterialMagenta);
		return;
	default:
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,
			MaterialCyan);
		return;
	}
	{
#ifdef WINVER
	struct tagColor {
		int red, green, blue;
	} color;

	color.red = GetRValue(w->rubik.faceGC[C]);
	color.green = GetGValue(w->rubik.faceGC[C]);
	color.blue = GetBValue(w->rubik.faceGC[C]);
#define MAXCOLOR 0xFF
#else
	XColor color;

	color.pixel = w->rubik.faceColor[C];
	XQueryColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
		&color);
#define MAXCOLOR 0xFFFF
#endif
#ifdef DEBUG
	(void) printf("i%d %d %d\n", color.red, color.green, color.blue);
#endif
	if (mono) {
		/* really GrayScale */
		float intensity = 0.3 * color.red + 0.59 * color.green +
			0.11 * color.blue;
#ifdef DEBUG
		(void) printf("m%g\n", intensity);
#endif
		Material[0] *= (float) intensity / MAXCOLOR;
		Material[1] *= (float) intensity / MAXCOLOR;
		Material[2] *= (float) intensity / MAXCOLOR;
	} else {
		Material[0] *= (float) color.red / MAXCOLOR;
		Material[1] *= (float) color.green / MAXCOLOR;
		Material[2] *= (float) color.blue / MAXCOLOR;
	}
	}
#ifdef DEBUG
	(void) printf("f%g %g %g\n", Material[0], Material[1], Material[2]);
#endif
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, Material);
}

static Boolean
draw_stickerless_cubit(void)
{
	glBegin(GL_QUADS);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
	/* Put sticker here */
	glNormal3f(0.0, 0.0, 1.0);
	glVertex3f(-CUBEROUND, -CUBEROUND, CUBELEN);
	glVertex3f(CUBEROUND, -CUBEROUND, CUBELEN);
	glVertex3f(CUBEROUND, CUBEROUND, CUBELEN);
	glVertex3f(-CUBEROUND, CUBEROUND, CUBELEN);
	glNormal3f(0.0, 0.0, -1.0);
	glVertex3f(-CUBEROUND, CUBEROUND, -CUBELEN);
	glVertex3f(CUBEROUND, CUBEROUND, -CUBELEN);
	glVertex3f(CUBEROUND, -CUBEROUND, -CUBELEN);
	glVertex3f(-CUBEROUND, -CUBEROUND, -CUBELEN);
	glNormal3f(-1.0, 0.0, 0.0);
	glVertex3f(-CUBELEN, -CUBEROUND, CUBEROUND);
	glVertex3f(-CUBELEN, CUBEROUND, CUBEROUND);
	glVertex3f(-CUBELEN, CUBEROUND, -CUBEROUND);
	glVertex3f(-CUBELEN, -CUBEROUND, -CUBEROUND);
	glNormal3f(1.0, 0.0, 0.0);
	glVertex3f(CUBELEN, -CUBEROUND, -CUBEROUND);
	glVertex3f(CUBELEN, CUBEROUND, -CUBEROUND);
	glVertex3f(CUBELEN, CUBEROUND, CUBEROUND);
	glVertex3f(CUBELEN, -CUBEROUND, CUBEROUND);
	glNormal3f(0.0, -1.0, 0.0);
	glVertex3f(CUBEROUND, -CUBELEN, -CUBEROUND);
	glVertex3f(CUBEROUND, -CUBELEN, CUBEROUND);
	glVertex3f(-CUBEROUND, -CUBELEN, CUBEROUND);
	glVertex3f(-CUBEROUND, -CUBELEN, -CUBEROUND);
	glNormal3f(0.0, 1.0, 0.0);
	glVertex3f(-CUBEROUND, CUBELEN, -CUBEROUND);
	glVertex3f(-CUBEROUND, CUBELEN, CUBEROUND);
	glVertex3f(CUBEROUND, CUBELEN, CUBEROUND);
	glVertex3f(CUBEROUND, CUBELEN, -CUBEROUND);

	/* Edges of cubit */
	glNormal3f(-1.0, -1.0, 0.0);
	glVertex3f(-CUBEROUND, -CUBELEN, -CUBEROUND);
	glVertex3f(-CUBEROUND, -CUBELEN, CUBEROUND);
	glVertex3f(-CUBELEN, -CUBEROUND, CUBEROUND);
	glVertex3f(-CUBELEN, -CUBEROUND, -CUBEROUND);
	glNormal3f(1.0, 1.0, 0.0);
	glVertex3f(CUBEROUND, CUBELEN, -CUBEROUND);
	glVertex3f(CUBEROUND, CUBELEN, CUBEROUND);
	glVertex3f(CUBELEN, CUBEROUND, CUBEROUND);
	glVertex3f(CUBELEN, CUBEROUND, -CUBEROUND);
	glNormal3f(-1.0, 1.0, 0.0);
	glVertex3f(-CUBELEN, CUBEROUND, -CUBEROUND);
	glVertex3f(-CUBELEN, CUBEROUND, CUBEROUND);
	glVertex3f(-CUBEROUND, CUBELEN, CUBEROUND);
	glVertex3f(-CUBEROUND, CUBELEN, -CUBEROUND);
	glNormal3f(1.0, -1.0, 0.0);
	glVertex3f(CUBELEN, -CUBEROUND, -CUBEROUND);
	glVertex3f(CUBELEN, -CUBEROUND, CUBEROUND);
	glVertex3f(CUBEROUND, -CUBELEN, CUBEROUND);
	glVertex3f(CUBEROUND, -CUBELEN, -CUBEROUND);
	glNormal3f(0.0, -1.0, -1.0);
	glVertex3f(-CUBEROUND, -CUBEROUND, -CUBELEN);
	glVertex3f(CUBEROUND, -CUBEROUND, -CUBELEN);
	glVertex3f(CUBEROUND, -CUBELEN, -CUBEROUND);
	glVertex3f(-CUBEROUND, -CUBELEN, -CUBEROUND);
	glNormal3f(0.0, 1.0, 1.0);
	glVertex3f(-CUBEROUND, CUBEROUND, CUBELEN);
	glVertex3f(CUBEROUND, CUBEROUND, CUBELEN);
	glVertex3f(CUBEROUND, CUBELEN, CUBEROUND);
	glVertex3f(-CUBEROUND, CUBELEN, CUBEROUND);
	glNormal3f(0.0, -1.0, 1.0);
	glVertex3f(-CUBEROUND, -CUBELEN, CUBEROUND);
	glVertex3f(CUBEROUND, -CUBELEN, CUBEROUND);
	glVertex3f(CUBEROUND, -CUBEROUND, CUBELEN);
	glVertex3f(-CUBEROUND, -CUBEROUND, CUBELEN);
	glNormal3f(0.0, 1.0, -1.0);
	glVertex3f(-CUBEROUND, CUBELEN, -CUBEROUND);
	glVertex3f(CUBEROUND, CUBELEN, -CUBEROUND);
	glVertex3f(CUBEROUND, CUBEROUND, -CUBELEN);
	glVertex3f(-CUBEROUND, CUBEROUND, -CUBELEN);
	glNormal3f(-1.0, 0.0, -1.0);
	glVertex3f(-CUBELEN, -CUBEROUND, -CUBEROUND);
	glVertex3f(-CUBELEN, CUBEROUND, -CUBEROUND);
	glVertex3f(-CUBEROUND, CUBEROUND, -CUBELEN);
	glVertex3f(-CUBEROUND, -CUBEROUND, -CUBELEN);
	glNormal3f(1.0, 0.0, 1.0);
	glVertex3f(CUBELEN, -CUBEROUND, CUBEROUND);
	glVertex3f(CUBELEN, CUBEROUND, CUBEROUND);
	glVertex3f(CUBEROUND, CUBEROUND, CUBELEN);
	glVertex3f(CUBEROUND, -CUBEROUND, CUBELEN);
	glNormal3f(1.0, 0.0, -1.0);
	glVertex3f(CUBEROUND, -CUBEROUND, -CUBELEN);
	glVertex3f(CUBEROUND, CUBEROUND, -CUBELEN);
	glVertex3f(CUBELEN, CUBEROUND, -CUBEROUND);
	glVertex3f(CUBELEN, -CUBEROUND, -CUBEROUND);
	glNormal3f(-1.0, 0.0, 1.0);
	glVertex3f(-CUBEROUND, -CUBEROUND, CUBELEN);
	glVertex3f(-CUBEROUND, CUBEROUND, CUBELEN);
	glVertex3f(-CUBELEN, CUBEROUND, CUBEROUND);
	glVertex3f(-CUBELEN, -CUBEROUND, CUBEROUND);
	glEnd();
	glBegin(GL_TRIANGLES);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
	/* Corners of cubit */
	glNormal3f(1.0, 1.0, 1.0);
	glVertex3f(CUBEROUND, CUBEROUND, CUBELEN);
	glVertex3f(CUBELEN, CUBEROUND, CUBEROUND);
	glVertex3f(CUBEROUND, CUBELEN, CUBEROUND);
	glNormal3f(-1.0, -1.0, -1.0);
	glVertex3f(-CUBEROUND, -CUBELEN, -CUBEROUND);
	glVertex3f(-CUBELEN, -CUBEROUND, -CUBEROUND);
	glVertex3f(-CUBEROUND, -CUBEROUND, -CUBELEN);
	glNormal3f(-1.0, 1.0, 1.0);
	glVertex3f(-CUBEROUND, CUBEROUND, CUBELEN);
	glVertex3f(-CUBEROUND, CUBELEN, CUBEROUND);
	glVertex3f(-CUBELEN, CUBEROUND, CUBEROUND);
	glNormal3f(1.0, -1.0, -1.0);
	glVertex3f(CUBELEN, -CUBEROUND, -CUBEROUND);
	glVertex3f(CUBEROUND, -CUBELEN, -CUBEROUND);
	glVertex3f(CUBEROUND, -CUBEROUND, -CUBELEN);
	glNormal3f(1.0, -1.0, 1.0);
	glVertex3f(CUBEROUND, -CUBEROUND, CUBELEN);
	glVertex3f(CUBEROUND, -CUBELEN, CUBEROUND);
	glVertex3f(CUBELEN, -CUBEROUND, CUBEROUND);
	glNormal3f(-1.0, 1.0, -1.0);
	glVertex3f(-CUBELEN, CUBEROUND, -CUBEROUND);
	glVertex3f(-CUBEROUND, CUBELEN, -CUBEROUND);
	glVertex3f(-CUBEROUND, CUBEROUND, -CUBELEN);
	glNormal3f(-1.0, -1.0, 1.0);
	glVertex3f(-CUBEROUND, -CUBEROUND, CUBELEN);
	glVertex3f(-CUBELEN, -CUBEROUND, CUBEROUND);
	glVertex3f(-CUBEROUND, -CUBELEN, CUBEROUND);
	glNormal3f(1.0, 1.0, -1.0);
	glVertex3f(CUBELEN, CUBEROUND, -CUBEROUND);
	glVertex3f(CUBEROUND, CUBEROUND, -CUBELEN);
	glVertex3f(CUBEROUND, CUBELEN, -CUBEROUND);
	glEnd();
	return True;
}

static Boolean
DrawCubit(RubikGLWidget w,
		int back, int front, int left, int right, int bottom, int top)
{
	Boolean mono = w->rubik.mono;
	char *buf1, *buf2;

	if (!draw_stickerless_cubit())
		return False;
	if (back != NO_POSITION) { /* Blue */
		glBegin(GL_POLYGON);
		pickColor(w, w->rubik.cubeLoc[BACK_FACE][back].face, mono);
		glNormal3f(0.0, 0.0, -1.0);
		glVertex3f(-STICKERSHORT, STICKERLONG, -STICKERDEPTH);
		glVertex3f(STICKERSHORT, STICKERLONG, -STICKERDEPTH);
		glVertex3f(STICKERLONG, STICKERSHORT, -STICKERDEPTH);
		glVertex3f(STICKERLONG, -STICKERSHORT, -STICKERDEPTH);
		glVertex3f(STICKERSHORT, -STICKERLONG, -STICKERDEPTH);
		glVertex3f(-STICKERSHORT, -STICKERLONG, -STICKERDEPTH);
		glVertex3f(-STICKERLONG, -STICKERSHORT, -STICKERDEPTH);
		glVertex3f(-STICKERLONG, STICKERSHORT, -STICKERDEPTH);
		glEnd();
		if (w->rubik.orient) {
			int rotation = w->rubik.cubeLoc[BACK_FACE][back].rotation;

			glBegin(GL_QUADS);
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
			glNormal3f(0.0, 0.0, -1.0);
			switch (rotation) {
			case 0:
				glVertex3f(CUT, -CUT, -CUTDEPTH);
				glVertex3f(CUT, -STICKERLONG, -CUTDEPTH);
				glVertex3f(-CUT, -STICKERLONG, -CUTDEPTH);
				glVertex3f(-CUT, -CUT, -CUTDEPTH);
				break;
			case 1:
				glVertex3f(CUT, CUT, -CUTDEPTH);
				glVertex3f(STICKERLONG, CUT, -CUTDEPTH);
				glVertex3f(STICKERLONG, -CUT, -CUTDEPTH);
				glVertex3f(CUT, -CUT, -CUTDEPTH);
				break;
			case 2:
				glVertex3f(-CUT, CUT, -CUTDEPTH);
				glVertex3f(-CUT, STICKERLONG, -CUTDEPTH);
				glVertex3f(CUT, STICKERLONG, -CUTDEPTH);
				glVertex3f(CUT, CUT, -CUTDEPTH);
				break;
			case 3:
				glVertex3f(-CUT, -CUT, -CUTDEPTH);
				glVertex3f(-STICKERLONG, -CUT, -CUTDEPTH);
				glVertex3f(-STICKERLONG, CUT, -CUTDEPTH);
				glVertex3f(-CUT, CUT, -CUTDEPTH);
				break;
			default:
				intCat(&buf1, "DrawCubit: face ", BACK_FACE);
				stringCat(&buf2, buf1, ", rotation ");
				free(buf1);
				intCat(&buf1, buf2, rotation);
				free(buf2);
				DISPLAY_WARNING(buf1);
				free(buf1);
			}
			glEnd();
		}
	}
	if (front != NO_POSITION) { /* White */
		glBegin(GL_POLYGON);
		glNormal3f(0.0, 0.0, 1.0);
		pickColor(w, w->rubik.cubeLoc[FRONT_FACE][front].face, mono);
		glVertex3f(-STICKERSHORT, -STICKERLONG, STICKERDEPTH);
		glVertex3f(STICKERSHORT, -STICKERLONG, STICKERDEPTH);
		glVertex3f(STICKERLONG, -STICKERSHORT, STICKERDEPTH);
		glVertex3f(STICKERLONG, STICKERSHORT, STICKERDEPTH);
		glVertex3f(STICKERSHORT, STICKERLONG, STICKERDEPTH);
		glVertex3f(-STICKERSHORT, STICKERLONG, STICKERDEPTH);
		glVertex3f(-STICKERLONG, STICKERSHORT, STICKERDEPTH);
		glVertex3f(-STICKERLONG, -STICKERSHORT, STICKERDEPTH);
		glEnd();
		if (w->rubik.orient) {
			int rotation = w->rubik.cubeLoc[FRONT_FACE][front].rotation;

			glBegin(GL_QUADS);
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
			glNormal3f(0.0, 0.0, 1.0);
			switch (rotation) {
			case 0:
				glVertex3f(CUT, CUT, CUTDEPTH);
				glVertex3f(CUT, STICKERLONG, CUTDEPTH);
				glVertex3f(-CUT, STICKERLONG, CUTDEPTH);
				glVertex3f(-CUT, CUT, CUTDEPTH);
				break;
			case 1:
				glVertex3f(CUT, -CUT, CUTDEPTH);
				glVertex3f(STICKERLONG, -CUT, CUTDEPTH);
				glVertex3f(STICKERLONG, CUT, CUTDEPTH);
				glVertex3f(CUT, CUT, CUTDEPTH);
				break;
			case 2:
				glVertex3f(-CUT, -CUT, CUTDEPTH);
				glVertex3f(-CUT, -STICKERLONG, CUTDEPTH);
				glVertex3f(CUT, -STICKERLONG, CUTDEPTH);
				glVertex3f(CUT, -CUT, CUTDEPTH);
				break;
			case 3:
				glVertex3f(-CUT, CUT, CUTDEPTH);
				glVertex3f(-STICKERLONG, CUT, CUTDEPTH);
				glVertex3f(-STICKERLONG, -CUT, CUTDEPTH);
				glVertex3f(-CUT, -CUT, CUTDEPTH);
				break;
			default:
				intCat(&buf1, "DrawCubit: face ", FRONT_FACE);
				stringCat(&buf2, buf1, ", rotation ");
				free(buf1);
				intCat(&buf1, buf2, rotation);
				free(buf2);
				DISPLAY_WARNING(buf1);
				free(buf1);
			}
			glEnd();
		}
	}
	if (left != NO_POSITION) { /* Yellow */
		glBegin(GL_POLYGON);
		pickColor(w, w->rubik.cubeLoc[LEFT_FACE][left].face, mono);
		glNormal3f(-1.0, 0.0, 0.0);
		glVertex3f(-STICKERDEPTH, -STICKERSHORT, STICKERLONG);
		glVertex3f(-STICKERDEPTH, STICKERSHORT, STICKERLONG);
		glVertex3f(-STICKERDEPTH, STICKERLONG, STICKERSHORT);
		glVertex3f(-STICKERDEPTH, STICKERLONG, -STICKERSHORT);
		glVertex3f(-STICKERDEPTH, STICKERSHORT, -STICKERLONG);
		glVertex3f(-STICKERDEPTH, -STICKERSHORT, -STICKERLONG);
		glVertex3f(-STICKERDEPTH, -STICKERLONG, -STICKERSHORT);
		glVertex3f(-STICKERDEPTH, -STICKERLONG, STICKERSHORT);
		glEnd();
		if (w->rubik.orient) {
			int rotation = w->rubik.cubeLoc[LEFT_FACE][left].rotation;

			glBegin(GL_QUADS);
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
			glNormal3f(-1.0, 0.0, 0.0);
			switch (rotation) {
			case 0:
				glVertex3f(-CUTDEPTH, CUT, CUT);
				glVertex3f(-CUTDEPTH, STICKERLONG, CUT);
				glVertex3f(-CUTDEPTH, STICKERLONG, -CUT);
				glVertex3f(-CUTDEPTH, CUT, -CUT);
				break;
			case 1:
				glVertex3f(-CUTDEPTH, -CUT, CUT);
				glVertex3f(-CUTDEPTH, -CUT, STICKERLONG);
				glVertex3f(-CUTDEPTH, CUT, STICKERLONG);
				glVertex3f(-CUTDEPTH, CUT, CUT);
				break;
			case 2:
				glVertex3f(-CUTDEPTH, -CUT, -CUT);
				glVertex3f(-CUTDEPTH, -STICKERLONG, -CUT);
				glVertex3f(-CUTDEPTH, -STICKERLONG, CUT);
				glVertex3f(-CUTDEPTH, -CUT, CUT);
				break;
			case 3:
				glVertex3f(-CUTDEPTH, CUT, -CUT);
				glVertex3f(-CUTDEPTH, CUT, -STICKERLONG);
				glVertex3f(-CUTDEPTH, -CUT, -STICKERLONG);
				glVertex3f(-CUTDEPTH, -CUT, -CUT);
				break;
			default:
				intCat(&buf1, "DrawCubit: face ", LEFT_FACE);
				stringCat(&buf2, buf1, ", rotation ");
				free(buf1);
				intCat(&buf1, buf2, rotation);
				free(buf2);
				DISPLAY_WARNING(buf1);
				free(buf1);
			}
			glEnd();
		}
	}
	if (right != NO_POSITION) { /* Green */
		glBegin(GL_POLYGON);
		pickColor(w, w->rubik.cubeLoc[RIGHT_FACE][right].face, mono);
		glNormal3f(1.0, 0.0, 0.0);
		glVertex3f(STICKERDEPTH, -STICKERSHORT, -STICKERLONG);
		glVertex3f(STICKERDEPTH, STICKERSHORT, -STICKERLONG);
		glVertex3f(STICKERDEPTH, STICKERLONG, -STICKERSHORT);
		glVertex3f(STICKERDEPTH, STICKERLONG, STICKERSHORT);
		glVertex3f(STICKERDEPTH, STICKERSHORT, STICKERLONG);
		glVertex3f(STICKERDEPTH, -STICKERSHORT, STICKERLONG);
		glVertex3f(STICKERDEPTH, -STICKERLONG, STICKERSHORT);
		glVertex3f(STICKERDEPTH, -STICKERLONG, -STICKERSHORT);
		glEnd();
		if (w->rubik.orient) {
			int rotation = w->rubik.cubeLoc[RIGHT_FACE][right].rotation;

			glBegin(GL_QUADS);
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
			glNormal3f(1.0, 0.0, 0.0);
			switch (rotation) {
			case 0:
				glVertex3f(CUTDEPTH, CUT, -CUT);
				glVertex3f(CUTDEPTH, STICKERLONG, -CUT);
				glVertex3f(CUTDEPTH, STICKERLONG, CUT);
				glVertex3f(CUTDEPTH, CUT, CUT);
				break;
			case 1:
				glVertex3f(CUTDEPTH, -CUT, -CUT);
				glVertex3f(CUTDEPTH, -CUT, -STICKERLONG);
				glVertex3f(CUTDEPTH, CUT, -STICKERLONG);
				glVertex3f(CUTDEPTH, CUT, -CUT);
				break;
			case 2:
				glVertex3f(CUTDEPTH, -CUT, CUT);
				glVertex3f(CUTDEPTH, -STICKERLONG, CUT);
				glVertex3f(CUTDEPTH, -STICKERLONG, -CUT);
				glVertex3f(CUTDEPTH, -CUT, -CUT);
				break;
			case 3:
				glVertex3f(CUTDEPTH, CUT, CUT);
				glVertex3f(CUTDEPTH, CUT, STICKERLONG);
				glVertex3f(CUTDEPTH, -CUT, STICKERLONG);
				glVertex3f(CUTDEPTH, -CUT, CUT);
				break;
			default:
				intCat(&buf1, "DrawCubit: face ", RIGHT_FACE);
				stringCat(&buf2, buf1, ", rotation ");
				free(buf1);
				intCat(&buf1, buf2, rotation);
				free(buf2);
				DISPLAY_WARNING(buf1);
				free(buf1);
			}
			glEnd();
		}
	}
	if (bottom != NO_POSITION) { /* Orange */
		glBegin(GL_POLYGON);
		pickColor(w, w->rubik.cubeLoc[BOTTOM_FACE][bottom].face, mono);
		glNormal3f(0.0, -1.0, 0.0);
		glVertex3f(STICKERLONG, -STICKERDEPTH, -STICKERSHORT);
		glVertex3f(STICKERLONG, -STICKERDEPTH, STICKERSHORT);
		glVertex3f(STICKERSHORT, -STICKERDEPTH, STICKERLONG);
		glVertex3f(-STICKERSHORT, -STICKERDEPTH, STICKERLONG);
		glVertex3f(-STICKERLONG, -STICKERDEPTH, STICKERSHORT);
		glVertex3f(-STICKERLONG, -STICKERDEPTH, -STICKERSHORT);
		glVertex3f(-STICKERSHORT, -STICKERDEPTH, -STICKERLONG);
		glVertex3f(STICKERSHORT, -STICKERDEPTH, -STICKERLONG);
		glEnd();
		if (w->rubik.orient) {
			int rotation = w->rubik.cubeLoc[BOTTOM_FACE][bottom].rotation;

			glBegin(GL_QUADS);
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
			glNormal3f(0.0, -1.0, 0.0);
			switch (rotation) {
			case 0:
				glVertex3f(CUT, -CUTDEPTH, CUT);
				glVertex3f(CUT, -CUTDEPTH, STICKERLONG);
				glVertex3f(-CUT, -CUTDEPTH, STICKERLONG);
				glVertex3f(-CUT, -CUTDEPTH, CUT);
				break;
			case 1:
				glVertex3f(CUT, -CUTDEPTH, -CUT);
				glVertex3f(STICKERLONG, -CUTDEPTH, -CUT);
				glVertex3f(STICKERLONG, -CUTDEPTH, CUT);
				glVertex3f(CUT, -CUTDEPTH, CUT);
				break;
			case 2:
				glVertex3f(-CUT, -CUTDEPTH, -CUT);
				glVertex3f(-CUT, -CUTDEPTH, -STICKERLONG);
				glVertex3f(CUT, -CUTDEPTH, -STICKERLONG);
				glVertex3f(CUT, -CUTDEPTH, -CUT);
				break;
			case 3:
				glVertex3f(-CUT, -CUTDEPTH, CUT);
				glVertex3f(-STICKERLONG, -CUTDEPTH, CUT);
				glVertex3f(-STICKERLONG, -CUTDEPTH, -CUT);
				glVertex3f(-CUT, -CUTDEPTH, -CUT);
				break;
			default:
				intCat(&buf1, "DrawCubit: face ", BOTTOM_FACE);
				stringCat(&buf2, buf1, ", rotation ");
				free(buf1);
				intCat(&buf1, buf2, rotation);
				free(buf2);
				DISPLAY_WARNING(buf1);
				free(buf1);
			}
			glEnd();
		}
	}
	if (top != NO_POSITION) { /* Red */
		glBegin(GL_POLYGON);
		pickColor(w, w->rubik.cubeLoc[TOP_FACE][top].face, mono);
		glNormal3f(0.0, 1.0, 0.0);
		glVertex3f(-STICKERLONG, STICKERDEPTH, -STICKERSHORT);
		glVertex3f(-STICKERLONG, STICKERDEPTH, STICKERSHORT);
		glVertex3f(-STICKERSHORT, STICKERDEPTH, STICKERLONG);
		glVertex3f(STICKERSHORT, STICKERDEPTH, STICKERLONG);
		glVertex3f(STICKERLONG, STICKERDEPTH, STICKERSHORT);
		glVertex3f(STICKERLONG, STICKERDEPTH, -STICKERSHORT);
		glVertex3f(STICKERSHORT, STICKERDEPTH, -STICKERLONG);
		glVertex3f(-STICKERSHORT, STICKERDEPTH, -STICKERLONG);
		glEnd();
		if (w->rubik.orient) {
			int rotation = w->rubik.cubeLoc[TOP_FACE][top].rotation;

			glBegin(GL_QUADS);
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MaterialGray);
			glNormal3f(0.0, 1.0, 0.0);
			switch (rotation) {
			case 0:
				glVertex3f(CUT, CUTDEPTH, -CUT);
				glVertex3f(CUT, CUTDEPTH, -STICKERLONG);
				glVertex3f(-CUT, CUTDEPTH, -STICKERLONG);
				glVertex3f(-CUT, CUTDEPTH, -CUT);
				break;
			case 1:
				glVertex3f(CUT, CUTDEPTH, CUT);
				glVertex3f(STICKERLONG, CUTDEPTH, CUT);
				glVertex3f(STICKERLONG, CUTDEPTH, -CUT);
				glVertex3f(CUT, CUTDEPTH, -CUT);
				break;
			case 2:
				glVertex3f(-CUT, CUTDEPTH, CUT);
				glVertex3f(-CUT, CUTDEPTH, STICKERLONG);
				glVertex3f(CUT, CUTDEPTH, STICKERLONG);
				glVertex3f(CUT, CUTDEPTH, CUT);
				break;
			case 3:
				glVertex3f(-CUT, CUTDEPTH, -CUT);
				glVertex3f(-STICKERLONG, CUTDEPTH, -CUT);
				glVertex3f(-STICKERLONG, CUTDEPTH, CUT);
				glVertex3f(-CUT, CUTDEPTH, CUT);
				break;
			default:
				intCat(&buf1, "DrawCubit: face ", TOP_FACE);
				stringCat(&buf2, buf1, ", rotation ");
				free(buf1);
				intCat(&buf1, buf2, rotation);
				free(buf2);
				DISPLAY_WARNING(buf1);
				free(buf1);
			}
			glEnd();
		}
	}
	return True;
}

/* Convert move to weird general notation */
static void
convertMove(RubikGLWidget w, RubikMove move, RubikSlice * slice)
{
	RubikLoc plane;
	int sizeOfRow, sizeOfColumn;

	plane = rotateSlice[(int) move.face][move.direction % 2];
	(*slice).face = plane.face;
	(*slice).rotation = plane.rotation;
	faceSizes((RubikWidget) w, move.face, &sizeOfRow, &sizeOfColumn);
	if (plane.face == 1 || /* VERTICAL */
			(plane.face == 2 &&
			(move.face == 1 || move.face == 3))) {
		if ((*slice).rotation == CW)
			(*slice).depth = sizeOfRow - 1 - move.position %
				sizeOfRow;
		else
			(*slice).depth = move.position % sizeOfRow;
	} else { /* (plane.face == 0 || *//* HORIZONTAL *//*
			(plane.face == 2 &&
			(move.face == 0 || move.face == 4))) */
		if ((*slice).rotation == CW)
			(*slice).depth = sizeOfColumn - 1 - move.position /
				sizeOfRow;
		else
			(*slice).depth = move.position / sizeOfRow;
	}
	/* If (*slice).depth = 0 then face 0, face 1, or face 2 moves */
	if (move.direction / 2)
		(*slice).rotation = ((*slice).rotation == CW) ? CCW : CW;
}

/* Assume the size is at least 2, or its just not challenging... */
static Boolean
DrawCube(RubikGLWidget w)
{
#define S1 1
#define SX ((GLint)S1*(MAXSIZEX-1))
#define SY ((GLint)S1*(MAXSIZEY-1))
#define SZ ((GLint)S1*(MAXSIZEZ-1))
#define HALFX (((GLfloat)MAXSIZEX-1.0)/2.0)
#define HALFY (((GLfloat)MAXSIZEY-1.0)/2.0)
#define HALFZ (((GLfloat)MAXSIZEZ-1.0)/2.0)
#define MIDX(a) (((GLfloat)(2*a-MAXSIZEX+1))/2.0)
#define MIDY(a) (((GLfloat)(2*a-MAXSIZEY+1))/2.0)
#define MIDZ(a) (((GLfloat)(2*a-MAXSIZEZ+1))/2.0)
#define DRAW_CUBIT(w,b,f,l,r,bm,t) if (!DrawCubit(w,b,f,l,r,bm,t)) return False
	RubikSlice slice = {NO_FACE, NO_ROTATION, NO_DEPTH};
	GLfloat rotateStep;
	int i, j, k;

	if (w->rubikGL.movement.face == NO_FACE) {
		slice.face = NO_FACE;
		slice.rotation = NO_ROTATION;
		slice.depth = NO_DEPTH;
	} else if (w->rubikGL.movement.face >= 0 && 
			w->rubikGL.movement.face < MAXFACES) {
		convertMove(w, w->rubikGL.movement, &slice);
	}
	rotateStep = (slice.rotation == CCW) ? w->rubikGL.rotateStep :
		-w->rubikGL.rotateStep;

/*-
 * The glRotatef() routine transforms the coordinate system for every future
 * vertex specification (this is not so simple, but by now comprehending this
 * is sufficient). So if you want to rotate the inner slice, you can draw
 * one slice, rotate the anglestep for the centerslice, draw the inner slice,
 * rotate reversely and draw the other slice.
 * There is a sequence for drawing cubies for each axis being moved...
 */
#if 0
slice.face = 0;
slice.face = FRONT_FACE;
(void) printf("slice.face %d\n", slice.face);
#endif
	switch (slice.face) {
	case NO_FACE:
	case TOP_FACE:	/* BOTTOM_FACE too */
		glPushMatrix();
		if (MAXSIZEY > 1 && slice.depth == MAXSIZEY - 1)
			glRotatef(rotateStep, 0, HALFY, 0);

		glTranslatef(-HALFX, -HALFY, -HALFZ);
		DRAW_CUBIT(w,
			FIRSTX + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
			FIRSTX + MAXSIZEX * LASTY : NO_POSITION,
			FIRSTZ + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
			LASTZ + MAXSIZEZ * LASTY : NO_POSITION,
			FIRSTX + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
			FIRSTX + MAXSIZEX * FIRSTZ : NO_POSITION);
		for (k = 1; k < MAXSIZEZ - 1; k++) {
			glTranslatef(0, 0, S1);
			DRAW_CUBIT(w,
				NO_POSITION, NO_POSITION,
				k + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
				REVZ(k) + MAXSIZEZ * LASTY : NO_POSITION,
				FIRSTX + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
				FIRSTX + MAXSIZEX * k : NO_POSITION);
		}
		if (MAXSIZEZ > 1) {
			glTranslatef(0, 0, S1);
			DRAW_CUBIT(w,
				NO_POSITION, FIRSTX + MAXSIZEX * LASTY,
				LASTZ + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
				FIRSTZ + MAXSIZEZ * LASTY : NO_POSITION,
				FIRSTX + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
				FIRSTX + MAXSIZEX * LASTZ : NO_POSITION);
			}
			for (i = 1; i < MAXSIZEX - 1; i++) {
				glTranslatef(S1, 0, -SZ);
				DRAW_CUBIT(w,
					i + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
					i + MAXSIZEX * LASTY : NO_POSITION,
					NO_POSITION, NO_POSITION,
					i + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
					i + MAXSIZEX * FIRSTZ : NO_POSITION);
				for (k = 1; k < MAXSIZEZ - 1; k++) {
					glTranslatef(0, 0, S1);
					DRAW_CUBIT(w,
						NO_POSITION, NO_POSITION,
						NO_POSITION, NO_POSITION,
						i + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
						i + MAXSIZEX * k : NO_POSITION);
				}
				if (MAXSIZEZ > 1) {
				glTranslatef(0, 0, S1);
				DRAW_CUBIT(w,
					NO_POSITION, i + MAXSIZEX * LASTY,
					NO_POSITION, NO_POSITION,
					i + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
					i + MAXSIZEX * LASTZ : NO_POSITION);
				}
			}
			if (MAXSIZEX > 1) {
			glTranslatef(S1, 0, -SZ);
			DRAW_CUBIT(w,
				LASTX + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
				LASTX + MAXSIZEX * LASTY : NO_POSITION,
				NO_POSITION, LASTZ + MAXSIZEZ * LASTY,
				LASTX + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
				LASTX + MAXSIZEX * FIRSTZ : NO_POSITION);
			if (MAXSIZEZ > 1) {
			for (k = 1; k < MAXSIZEZ - 1; k++) {
				glTranslatef(0, 0, S1);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, REVZ(k) + MAXSIZEZ * LASTY,
					LASTX + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
					LASTX + MAXSIZEX * k : NO_POSITION);
			}
			glTranslatef(0, 0, S1);
			DRAW_CUBIT(w,
				NO_POSITION, LASTX + MAXSIZEX * LASTY,
				NO_POSITION, FIRSTZ + MAXSIZEZ * LASTY,
				LASTX + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
				LASTX + MAXSIZEX * LASTZ : NO_POSITION);
			}
			}
			glPopMatrix();
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glPushMatrix();
				if (slice.depth == REVY(j))
					glRotatef(rotateStep, 0, HALFY, 0);
				glTranslatef(-HALFX, MIDY(j), -HALFZ);
				DRAW_CUBIT(w,
					FIRSTX + MAXSIZEX * j, (MAXSIZEZ == 1) ?
					FIRSTX + MAXSIZEX * REVY(j) : NO_POSITION,
					FIRSTZ + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
					LASTZ + MAXSIZEZ * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION);
				for (k = 1; k < MAXSIZEZ - 1; k++) {
					glTranslatef(0, 0, S1);
					DRAW_CUBIT(w,
						NO_POSITION, NO_POSITION,
						k + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
						REVZ(k) + MAXSIZEZ * REVY(j) : NO_POSITION,
						NO_POSITION, NO_POSITION);
				}
				if (MAXSIZEZ > 1) {
				glTranslatef(0, 0, S1);
				DRAW_CUBIT(w,
					NO_POSITION, FIRSTX + MAXSIZEX * REVY(j),
					LASTZ + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
					FIRSTZ + MAXSIZEZ * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION);
				}
				for (i = 1; i < MAXSIZEX - 1; i++) {
					glTranslatef(1, 0, -SZ);
					DRAW_CUBIT(w,
						i + MAXSIZEX * j, (MAXSIZEZ == 1) ?
						i + MAXSIZEX * REVY(j) : NO_POSITION,
						NO_POSITION, NO_POSITION,
						NO_POSITION, NO_POSITION);
					/* Center */
					if (MAXSIZEZ > 1) {
					glTranslatef(0, 0, SZ);
					DRAW_CUBIT(w,
						NO_POSITION, i + MAXSIZEX * REVY(j),
						NO_POSITION, NO_POSITION,
						NO_POSITION, NO_POSITION);
					}
				}
				if (MAXSIZEX > 1) {
				glTranslatef(S1, 0, -SZ);
				DRAW_CUBIT(w,
					LASTX + MAXSIZEX * j, (MAXSIZEZ == 1) ?
					LASTX + MAXSIZEX * REVY(j) : NO_POSITION,
					NO_POSITION, LASTZ + MAXSIZEZ * REVY(j),
					NO_POSITION, NO_POSITION);
				for (k = 1; k < MAXSIZEZ - 1; k++) {
					glTranslatef(0, 0, S1);
					DRAW_CUBIT(w,
						NO_POSITION, NO_POSITION,
						NO_POSITION, REVZ(k) + MAXSIZEZ * REVY(j),
						NO_POSITION, NO_POSITION);
				}
				if (MAXSIZEZ > 1) {
				glTranslatef(0, 0, S1);
				DRAW_CUBIT(w,
					NO_POSITION, LASTX + MAXSIZEX * REVY(j),
					NO_POSITION, FIRSTZ + MAXSIZEZ * REVY(j),
					NO_POSITION, NO_POSITION);
				}
				}
				glPopMatrix();
			}
			if (MAXSIZEY > 1) {
			if (slice.depth == 0)
				glRotatef(rotateStep, 0, HALFY, 0);
			glTranslatef(-HALFX, HALFY, -HALFZ);
			DRAW_CUBIT(w,
				FIRSTX + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
				FIRSTX + MAXSIZEX * FIRSTY : NO_POSITION,
				FIRSTZ + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
				LASTZ + MAXSIZEZ * FIRSTY : NO_POSITION,
				NO_POSITION, FIRSTX + MAXSIZEX * FIRSTZ);
			for (k = 1; k < MAXSIZEZ - 1; k++) {
				glTranslatef(0, 0, S1);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					k + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
					REVZ(k) + MAXSIZEZ * FIRSTY : NO_POSITION,
					NO_POSITION, FIRSTX + MAXSIZEX * k);
			}
			if (MAXSIZEZ > 1) {
			glTranslatef(0, 0, S1);
			DRAW_CUBIT(w,
				NO_POSITION, FIRSTX + MAXSIZEX * FIRSTY,
				LASTZ + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
				FIRSTZ + MAXSIZEZ * FIRSTY : NO_POSITION,
				NO_POSITION, FIRSTX + MAXSIZEX * LASTZ);
			}
			for (i = 1; i < MAXSIZEX - 1; i++) {
				glTranslatef(S1, 0, -SZ);
				DRAW_CUBIT(w,
					i + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
					i + MAXSIZEX * FIRSTY : NO_POSITION,
					NO_POSITION, NO_POSITION,
					NO_POSITION, i + MAXSIZEX * FIRSTZ);
				for (k = 1; k < MAXSIZEZ - 1; k++) {
					glTranslatef(0, 0, S1);
					DRAW_CUBIT(w,
						NO_POSITION, NO_POSITION,
						NO_POSITION, NO_POSITION,
						NO_POSITION, i + MAXSIZEX * k);
				}
				if (MAXSIZEZ > 1) {
				glTranslatef(0, 0, S1);
				DRAW_CUBIT(w,
					NO_POSITION, i + MAXSIZEX * FIRSTY,
					NO_POSITION, NO_POSITION,
					NO_POSITION, i + MAXSIZEX * LASTZ);
				}
			}
			if (MAXSIZEX > 1) {
			glTranslatef(S1, 0, -SZ);
			DRAW_CUBIT(w,
				LASTX + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
				LASTX + MAXSIZEX * FIRSTY : NO_POSITION,
				NO_POSITION, LASTZ + MAXSIZEZ * FIRSTY,
				NO_POSITION, LASTX + MAXSIZEX * FIRSTZ);
			for (k = 1; k < MAXSIZEZ - 1; k++) {
				glTranslatef(0, 0, S1);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, REVZ(k) + MAXSIZEZ * FIRSTY,
					NO_POSITION, LASTX + MAXSIZEX * k);
			}
			if (MAXSIZEZ > 1) {
			glTranslatef(0, 0, S1);
			DRAW_CUBIT(w,
				NO_POSITION, LASTX + MAXSIZEX * FIRSTY,
				NO_POSITION, FIRSTZ + MAXSIZEZ * FIRSTY,
				NO_POSITION, LASTX + MAXSIZEX * LASTZ);
			}
			}
		}
		break;
	case LEFT_FACE:	/* RIGHT_FACE too */
		/* rotateStep is negative because the RIGHT face is the default here */
		glPushMatrix();
		if (MAXSIZEX > 1 && slice.depth == 0)
			glRotatef(-rotateStep, HALFX, 0, 0);

		glTranslatef(-HALFX, -HALFY, -HALFZ);
		DRAW_CUBIT(w,
			FIRSTX + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
			FIRSTX + MAXSIZEX * LASTY : NO_POSITION,
			FIRSTZ + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
			LASTZ + MAXSIZEZ * LASTY : NO_POSITION,
			FIRSTX + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
			FIRSTX + MAXSIZEX * FIRSTZ : NO_POSITION);
		for (j = 1; j < MAXSIZEY - 1; j++) {
			glTranslatef(0, S1, 0);
			DRAW_CUBIT(w,
				FIRSTX + MAXSIZEX * j, (MAXSIZEZ == 1) ?
				FIRSTX + MAXSIZEX * REVY(j) : NO_POSITION,
				FIRSTZ + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
				LASTZ + MAXSIZEZ * REVY(j) : NO_POSITION,
				NO_POSITION, NO_POSITION);
		}
		if (MAXSIZEY > 1) {
		glTranslatef(0, S1, 0);
		DRAW_CUBIT(w,
			FIRSTX + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
			FIRSTX + MAXSIZEX * FIRSTY : NO_POSITION,
			FIRSTZ + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
			LASTZ + MAXSIZEZ * FIRSTY : NO_POSITION,
			NO_POSITION, FIRSTX + MAXSIZEX * FIRSTZ);
		}
		for (k = 1; k < MAXSIZEZ - 1; k++) {
			glTranslatef(0, -SY, S1);
			DRAW_CUBIT(w,
				NO_POSITION, NO_POSITION,
				k + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
				REVZ(k) + MAXSIZEZ * LASTY : NO_POSITION,
				FIRSTX + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
				FIRSTX + MAXSIZEX * k : NO_POSITION);
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					k + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
					REVZ(k) + MAXSIZEZ * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION);
			}
			if (MAXSIZEY > 1) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					k + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
					REVZ(k) + MAXSIZEZ * FIRSTY : NO_POSITION,
					NO_POSITION, FIRSTX + MAXSIZEX * k);
			}
		}
		if (MAXSIZEZ > 1) {
		glTranslatef(0, -SY, S1);
		DRAW_CUBIT(w,
			NO_POSITION, FIRSTX + MAXSIZEX * LASTY,
			LASTZ + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
			FIRSTZ + MAXSIZEZ * LASTY : NO_POSITION,
			FIRSTX + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
				FIRSTX + MAXSIZEX * LASTZ : NO_POSITION);
			if (MAXSIZEY > 1) {
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, FIRSTX + MAXSIZEX * REVY(j),
					LASTZ + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
					FIRSTZ + MAXSIZEZ * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION);
			}
			glTranslatef(0, S1, 0);
			DRAW_CUBIT(w,
				NO_POSITION, FIRSTX + MAXSIZEX * FIRSTY,
				LASTZ + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
				FIRSTZ + MAXSIZEZ * FIRSTY : NO_POSITION,
				NO_POSITION, FIRSTX + MAXSIZEX * LASTZ);
			}
		}
		glPopMatrix();
		for (i = 1; i < MAXSIZEX - 1; i++) {
			glPushMatrix();
			if (slice.depth == i)
				glRotatef(-rotateStep, HALFX, 0, 0);
			glTranslatef(MIDX(i), -HALFY, -HALFZ);
			DRAW_CUBIT(w,
				i + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
				i + MAXSIZEX * LASTY : NO_POSITION,
				NO_POSITION, NO_POSITION,
				i + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
				i + MAXSIZEX * FIRSTZ : NO_POSITION);
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					i + MAXSIZEX * j, (MAXSIZEZ == 1) ?
					i + MAXSIZEX * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION,
					NO_POSITION, NO_POSITION);
			}
			if (MAXSIZEY > 1) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					i + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
					i + MAXSIZEX * FIRSTY : NO_POSITION,
					NO_POSITION, NO_POSITION,
					NO_POSITION, i + MAXSIZEX * FIRSTZ);
			}
			for (k = 1; k < MAXSIZEZ - 1; k++) {
				glTranslatef(0, -SY, S1);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, NO_POSITION,
					i + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
					i + MAXSIZEX * k : NO_POSITION);
				/* Center */
				if (MAXSIZEY > 1) {
				glTranslatef(0, SY, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, NO_POSITION,
					NO_POSITION, i + MAXSIZEX * k);
				}
			}
			if (MAXSIZEZ > 1) {
				glTranslatef(0, -SY, S1);
				DRAW_CUBIT(w,
					NO_POSITION, i + MAXSIZEX * LASTY,
					NO_POSITION, NO_POSITION,
					i + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
					i + MAXSIZEX * LASTZ : NO_POSITION);
				for (j = 1; j < MAXSIZEY - 1; j++) {
					glTranslatef(0, S1, 0);
					DRAW_CUBIT(w,
						NO_POSITION, i + MAXSIZEX * REVY(j),
						NO_POSITION, NO_POSITION,
						NO_POSITION, NO_POSITION);
				}
				if (MAXSIZEY > 1) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, i + MAXSIZEX * FIRSTY,
					NO_POSITION, NO_POSITION,
					NO_POSITION, i + MAXSIZEX * LASTZ);
				}
			}
			glPopMatrix();
		}
		if (MAXSIZEX > 1) {
			if (slice.depth == MAXSIZEX - 1)
				glRotatef(-rotateStep, HALFX, 0, 0);
			glTranslatef(HALFX, -HALFY, -HALFZ);
			DRAW_CUBIT(w,
				LASTX + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
				LASTX + MAXSIZEX * LASTY : NO_POSITION,
				NO_POSITION, LASTZ + MAXSIZEZ * LASTY,
				LASTX + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
				LASTX + MAXSIZEX * FIRSTZ : NO_POSITION);
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					LASTX + MAXSIZEX * j, (MAXSIZEZ == 1) ?
					LASTX + MAXSIZEX * REVY(j) : NO_POSITION,
					NO_POSITION, LASTZ + MAXSIZEZ * REVY(j),
					NO_POSITION, NO_POSITION);
			}
			if (MAXSIZEY > 1) {
			glTranslatef(0, S1, 0);
			DRAW_CUBIT(w,
				LASTX + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
				LASTX + MAXSIZEX * FIRSTY : NO_POSITION,
				NO_POSITION, LASTZ + MAXSIZEZ * FIRSTY,
				NO_POSITION, LASTX + MAXSIZEX * FIRSTZ);
			}
			for (k = 1; k < MAXSIZEZ - 1; k++) {
				glTranslatef(0, -SY, S1);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, REVZ(k) + MAXSIZEZ * LASTY,
					LASTX + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
					LASTX + MAXSIZEX * k : NO_POSITION);
				for (j = 1; j < MAXSIZEY - 1; j++) {
					glTranslatef(0, S1, 0);
					DRAW_CUBIT(w,
						NO_POSITION, NO_POSITION,
						NO_POSITION, REVZ(k) + MAXSIZEZ * REVY(j),
						NO_POSITION, NO_POSITION);
				}
				if (MAXSIZEY > 1) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, REVZ(k) + MAXSIZEZ * FIRSTY,
					NO_POSITION, LASTX + MAXSIZEX * k);
				}
			}
			if (MAXSIZEZ > 1) {
			glTranslatef(0, -SY, S1);
			DRAW_CUBIT(w,
				NO_POSITION, LASTX + MAXSIZEX * LASTY,
				NO_POSITION, FIRSTZ + MAXSIZEZ * LASTY,
				LASTX + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
				LASTX + MAXSIZEX * LASTZ : NO_POSITION);
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glTranslatef(0, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, LASTX + MAXSIZEX * REVY(j),
					NO_POSITION, FIRSTZ + MAXSIZEZ * REVY(j),
					NO_POSITION, NO_POSITION);
			}
			if (MAXSIZEY > 1) {
			glTranslatef(0, S1, 0);
			DRAW_CUBIT(w,
				NO_POSITION, LASTX + MAXSIZEX * FIRSTY,
				NO_POSITION, FIRSTZ + MAXSIZEZ * FIRSTY,
				NO_POSITION, LASTX + MAXSIZEX * LASTZ);
			}
			}
		}
		break;
	case FRONT_FACE:	/* BACK_FACE too */
		glPushMatrix();
		if (MAXSIZEZ > 1 && slice.depth == MAXSIZEZ - 1)
			glRotatef(rotateStep, 0, 0, HALFZ);

		glTranslatef(-HALFX, -HALFY, -HALFZ);
		DRAW_CUBIT(w,
			FIRSTX + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
			FIRSTX + MAXSIZEX * LASTY : NO_POSITION,
			FIRSTZ + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
			LASTZ + MAXSIZEZ * LASTY : NO_POSITION,
			FIRSTX + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
			FIRSTX + MAXSIZEX * FIRSTZ : NO_POSITION);
		for (i = 1; i < MAXSIZEX - 1; i++) {
			glTranslatef(S1, 0, 0);
			DRAW_CUBIT(w,
				i + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
				i + MAXSIZEX * LASTY : NO_POSITION,
				NO_POSITION, NO_POSITION,
				i + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
				i + MAXSIZEX * FIRSTZ : NO_POSITION);
		}
		if (MAXSIZEX > 1) {
			glTranslatef(S1, 0, 0);
			DRAW_CUBIT(w,
				LASTX + MAXSIZEX * FIRSTY, (MAXSIZEZ == 1) ?
				LASTX + MAXSIZEX * LASTY : NO_POSITION,
				NO_POSITION, LASTZ + MAXSIZEZ * LASTY,
				LASTX + MAXSIZEX * LASTZ, (MAXSIZEY == 1) ?
				LASTX + MAXSIZEX * FIRSTZ : NO_POSITION);
		}
		for (j = 1; j < MAXSIZEY - 1; j++) {
			glTranslatef(-SX, S1, 0);
			DRAW_CUBIT(w,
				FIRSTX + MAXSIZEX * j, (MAXSIZEZ == 1) ?
				FIRSTX + MAXSIZEX * REVY(j) : NO_POSITION,
				FIRSTZ + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
				LASTZ + MAXSIZEZ * REVY(j) : NO_POSITION,
				NO_POSITION, NO_POSITION);
			for (i = 1; i < MAXSIZEX - 1; i++) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					i + MAXSIZEX * j, (MAXSIZEZ == 1) ?
					i + MAXSIZEX * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION,
					NO_POSITION, NO_POSITION);
			}
			if (MAXSIZEX > 1) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					LASTX + MAXSIZEX * j, (MAXSIZEZ == 1) ?
					LASTX + MAXSIZEX * REVY(j) : NO_POSITION,
					NO_POSITION, LASTZ + MAXSIZEZ * REVY(j),
					NO_POSITION, NO_POSITION);
			}
		}
		if (MAXSIZEY > 1) {
			glTranslatef(-SX, S1, 0);
			DRAW_CUBIT(w,
				FIRSTX + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
				FIRSTX + MAXSIZEX * FIRSTY : NO_POSITION,
				FIRSTZ + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
				LASTZ + MAXSIZEZ * FIRSTY : NO_POSITION,
				NO_POSITION, FIRSTX + MAXSIZEX * FIRSTZ);
			if (MAXSIZEX > 1) {
			for (i = 1; i < MAXSIZEX - 1; i++) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					i + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
					i + MAXSIZEX * FIRSTY : NO_POSITION,
					NO_POSITION, NO_POSITION,
					NO_POSITION, i + MAXSIZEX * FIRSTZ);
			}
			glTranslatef(S1, 0, 0);
			DRAW_CUBIT(w,
				LASTX + MAXSIZEX * LASTY, (MAXSIZEZ == 1) ?
				LASTX + MAXSIZEX * FIRSTY : NO_POSITION,
				NO_POSITION, LASTZ + MAXSIZEZ * FIRSTY,
				NO_POSITION, LASTX + MAXSIZEX * FIRSTZ);
			}
		}
		glPopMatrix();
		for (k = 1; k < MAXSIZEZ - 1; k++) {
			glPushMatrix();
			if (slice.depth == REVZ(k))
				glRotatef(rotateStep, 0, 0, HALFZ);
			glTranslatef(-HALFX, -HALFY, MIDZ(k));
			DRAW_CUBIT(w,
				NO_POSITION, NO_POSITION,
				k + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
				REVZ(k) + MAXSIZEZ * LASTY : NO_POSITION,
				FIRSTX + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
				FIRSTX + MAXSIZEX * k : NO_POSITION);
			for (i = 1; i < MAXSIZEX - 1; i++) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, NO_POSITION,
					i + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
					i + MAXSIZEX * k : NO_POSITION);
			}
			if (MAXSIZEX > 1) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, REVZ(k) + MAXSIZEZ * LASTY,
					LASTX + MAXSIZEX * REVZ(k), (MAXSIZEY == 1) ?
					LASTX + MAXSIZEX * k : NO_POSITION);
			}
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glTranslatef(-SX, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					k + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
					REVZ(k) + MAXSIZEZ * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION);
				/* Center */
				if (MAXSIZEX > 1) {
				glTranslatef(SX, 0, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, REVZ(k) + MAXSIZEZ * REVY(j),
					NO_POSITION, NO_POSITION);
				}
			}
			if (MAXSIZEY > 1) {
				glTranslatef(-SX, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					k + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
					REVZ(k) + MAXSIZEZ * FIRSTY : NO_POSITION,
					NO_POSITION, FIRSTX + MAXSIZEX * k);
				for (i = 1; i < MAXSIZEX - 1; i++) {
					glTranslatef(S1, 0, 0);
					DRAW_CUBIT(w,
						NO_POSITION, NO_POSITION,
						NO_POSITION, NO_POSITION,
						NO_POSITION, i + MAXSIZEX * k);
				}
				if (MAXSIZEX > 1) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					NO_POSITION, NO_POSITION,
					NO_POSITION, REVZ(k) + MAXSIZEZ * FIRSTY,
					NO_POSITION, LASTX + MAXSIZEX * k);
				}
			}
			glPopMatrix();
		}
		if (MAXSIZEZ > 1) {
			if (slice.depth == 0)
				glRotatef(rotateStep, 0, 0, HALFZ);
			glTranslatef(-HALFX, -HALFY, HALFZ);
			DRAW_CUBIT(w,
				NO_POSITION, FIRSTX + MAXSIZEX * LASTY,
				LASTZ + MAXSIZEZ * LASTY, (MAXSIZEX == 1) ?
				FIRSTZ + MAXSIZEZ * LASTY : NO_POSITION,
				FIRSTX + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
				FIRSTX + MAXSIZEX * LASTZ : NO_POSITION);
			for (i = 1; i < MAXSIZEX - 1; i++) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					NO_POSITION, i + MAXSIZEX * LASTY,
					NO_POSITION, NO_POSITION,
					i + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
					i + MAXSIZEX * LASTZ : NO_POSITION);
			}
			if (MAXSIZEX > 1) {
			glTranslatef(S1, 0, 0);
			DRAW_CUBIT(w,
				NO_POSITION, LASTX + MAXSIZEX * LASTY,
				NO_POSITION, FIRSTZ + MAXSIZEZ * LASTY,
				LASTX + MAXSIZEX * FIRSTZ, (MAXSIZEY == 1) ?
				LASTX + MAXSIZEX * LASTZ : NO_POSITION);
			}
			for (j = 1; j < MAXSIZEY - 1; j++) {
				glTranslatef(-SX, S1, 0);
				DRAW_CUBIT(w,
					NO_POSITION, FIRSTX + MAXSIZEX * REVY(j),
					LASTZ + MAXSIZEZ * REVY(j), (MAXSIZEX == 1) ?
					FIRSTZ + MAXSIZEZ * REVY(j) : NO_POSITION,
					NO_POSITION, NO_POSITION);
				for (i = 1; i < MAXSIZEX - 1; i++) {
					glTranslatef(S1, 0, 0);
					DRAW_CUBIT(w,
						NO_POSITION, i + MAXSIZEX * REVY(j),
						NO_POSITION, NO_POSITION,
						NO_POSITION, NO_POSITION);
				}
				if (MAXSIZEX > 1) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					NO_POSITION, LASTX + MAXSIZEX * REVY(j),
					NO_POSITION, FIRSTZ + MAXSIZEZ * REVY(j),
					NO_POSITION, NO_POSITION);
				}
			}
			if (MAXSIZEY > 1) {
			glTranslatef(-SX, S1, 0);
			DRAW_CUBIT(w,
				NO_POSITION, FIRSTX + MAXSIZEX * FIRSTY,
				LASTZ + MAXSIZEZ * FIRSTY, (MAXSIZEX == 1) ?
				FIRSTZ + MAXSIZEZ * FIRSTY : NO_POSITION,
				NO_POSITION, FIRSTX + MAXSIZEX * LASTZ);
			for (i = 1; i < MAXSIZEX - 1; i++) {
				glTranslatef(S1, 0, 0);
				DRAW_CUBIT(w,
					NO_POSITION, i + MAXSIZEX * FIRSTY,
					NO_POSITION, NO_POSITION,
					NO_POSITION, i + MAXSIZEX * LASTZ);
			}
			if (MAXSIZEX > 1) {
			glTranslatef(S1, 0, 0);
			DRAW_CUBIT(w,
				NO_POSITION, LASTX + MAXSIZEX * FIRSTY,
				NO_POSITION, FIRSTZ + MAXSIZEZ * FIRSTY,
				NO_POSITION, LASTX + MAXSIZEX * LASTZ);
			}
			}
		}
		break;
	}
	return True;
#undef S1
}

void
MovePiecesGL(RubikGLWidget w, int face, int position, int direction,
		Boolean control, int fast)
{
	int i;

	w->rubikGL.movement.face = face;
	w->rubikGL.movement.position = position;
	w->rubikGL.movement.direction = direction;
	w->rubikGL.rotateStep = 0.0;
	w->rubikGL.angleStepX = 0.0;
	w->rubikGL.angleStepY = 0.0;
	w->rubikGL.angleStepZ = 0.0;
	if (control) {
		double angleStep = 0.0;
		if (w->rubik.degreeTurn == 90)
			for (i = 0; i < w->rubikGL.numTurnInc / fast; i++) {
				angleStep += 90.0 * fast / w->rubikGL.numTurnInc;
				if (rotateDir[face][direction] == -1)
					w->rubikGL.angleStepY = -angleStep;
				else if (rotateDir[face][direction] == 1)
					w->rubikGL.angleStepY = angleStep;
				else if (rotateDir[face][direction] == -2)
					w->rubikGL.angleStepX = -angleStep;
				else if (rotateDir[face][direction] == 2)
					w->rubikGL.angleStepX = angleStep;
				else if (rotateDir[face][direction] == -3)
					w->rubikGL.angleStepZ = -angleStep;
				else if (rotateDir[face][direction] == 3)
					w->rubikGL.angleStepZ = angleStep;
				DrawAllPiecesGL(w);
				Sleep(w->rubik.delay / (w->rubikGL.numTurnInc * fast));
			}
		else
			for (i = 0; i < w->rubikGL.numTurnInc * 2 / fast; i++) {
				angleStep += 90.0 * fast / w->rubikGL.numTurnInc;
				if (rotateDir[face][direction] == -1)
					w->rubikGL.angleStepY = -angleStep;
				else if (rotateDir[face][direction] == 1)
					w->rubikGL.angleStepY = angleStep;
				else if (rotateDir[face][direction] == -2)
					w->rubikGL.angleStepX = -angleStep;
				else if (rotateDir[face][direction] == 2)
					w->rubikGL.angleStepX = angleStep;
				else if (rotateDir[face][direction] == -3)
					w->rubikGL.angleStepZ = -angleStep;
				else if (rotateDir[face][direction] == 3)
					w->rubikGL.angleStepZ = angleStep;
				DrawAllPiecesGL(w);
				Sleep(w->rubik.delay / (w->rubikGL.numTurnInc * fast) *
					2 / 3); /* Longer so move a little faster */
			}
		w->rubikGL.angleStepX = 0.0;
		w->rubikGL.angleStepY = 0.0;
		w->rubikGL.angleStepZ = 0.0;
	} else {
		if (w->rubik.degreeTurn == 90)
			for (i = 0; i < w->rubikGL.numTurnInc / fast; i++) {
				w->rubikGL.rotateStep += 90.0 * fast / w->rubikGL.numTurnInc;
				DrawAllPiecesGL(w);
				Sleep(w->rubik.delay / (w->rubikGL.numTurnInc * fast));
			}
		else
			for (i = 0; i < w->rubikGL.numTurnInc * 2 / fast; i++) {
				w->rubikGL.rotateStep += 90.0 * fast / w->rubikGL.numTurnInc;
				DrawAllPiecesGL(w);
				Sleep(w->rubik.delay / (w->rubikGL.numTurnInc * fast) *
					2 / 3); /* Longer so move a little faster */
			}
	}
	w->rubikGL.movement.face = NO_FACE;
	w->rubikGL.movement.position = NO_POSITION;
	w->rubikGL.movement.direction = NO_DEPTH;
}

void
DrawAllPiecesGL(RubikGLWidget w)
{
#ifdef WINVER
	wglMakeCurrent(w->core.hDC, hRC);
#else
	if (!glXMakeCurrent(XtDisplay(w), XtWindow(w), *glXContext)) {
		DISPLAY_WARNING("Draw All GL error");
	}
#endif
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
	glTranslatef(0.0, 0.0, -10.0);
	if (w->core.height > w->core.width)
		glScalef(SCALE, SCALE * w->core.width / w->core.height, SCALE);
	else
		glScalef(SCALE * w->core.height / w->core.width, SCALE, SCALE);
	switch (w->rubik.view) {
	case 0: /* >-  vertical */
		glRotatef(w->rubikGL.step / 2, 0, 1, 0);
		glRotatef(w->rubikGL.step / 2, 1, 0, 0);
		break;
	case 1: /* Y  !vertical */
		glRotatef(w->rubikGL.step / 2, 1, 0, 0);
		glRotatef(w->rubikGL.step / 2, 0, 1, 0);
		break;
	case 2: /* -<  vertical */
		glRotatef(2 * w->rubikGL.step, 0, 0, 1);
		glRotatef(5 * w->rubikGL.step / 2, 0, 1, 0);
		glRotatef(w->rubikGL.step / 2, 1, 0, 0);
		break;
	case 3: /* ^  !vertical */
		glRotatef(5 * w->rubikGL.step / 2, 0, 1, 0);
		glRotatef(w->rubikGL.step / 2, 1, 0, 1);
		break;
	default:
		break;
	}
	glRotatef(w->rubikGL.angleStepX, 0, 1, 0);
	glRotatef(w->rubikGL.angleStepY, 1, 0, 0);
	glRotatef(w->rubikGL.angleStepZ, 0, 0, 1);
	if (!DrawCube(w)) {
		DISPLAY_WARNING("DrawCube error");
	}
	glPopMatrix();
	glFlush();
#ifdef WINVER
	SwapBuffers(w->core.hDC);
#else
	glXSwapBuffers(XtDisplay(w), XtWindow(w));
#endif
}

void
DrawFrameGL(RubikGLWidget w, Boolean focus)
{
	/* UNUSED */
}

Boolean
SelectPiecesGL(RubikGLWidget w, int x, int y, int *face, int *position)
{
	/* UNUSED */
	return True;
}

#ifdef WINVER
Boolean setupPixelFormat(RubikGLWidget w, BYTE type, DWORD flags)
{
	PIXELFORMATDESCRIPTOR pfd;
	int pixelFormat;

	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | flags;
	pfd.dwLayerMask = PFD_MAIN_PLANE;
	/*pfd.cColorBits = 8;
	pfd.cDepthBits = 16;*/
	pfd.iPixelType = type;
	pfd.cColorBits = 24;
	pfd.cDepthBits = 32;

	if ((pixelFormat = ChoosePixelFormat(w->core.hDC, &pfd)) == 0) {
		DWORD err;
		char *buf1;

		err = GetLastError();
		/* 6 ERROR_INVALID_HANDLE */
		intCat(&buf1, "ChoosePixelFormat failed ", err);
		DISPLAY_WARNING(buf1);
		free(buf1);
		return FALSE;
	}

	if (SetPixelFormat(w->core.hDC, pixelFormat, &pfd) == FALSE) {
		MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
		return FALSE;
	}
	DescribePixelFormat(w->core.hDC, pixelFormat,
		sizeof(PIXELFORMATDESCRIPTOR), &pfd);
	return TRUE;
}
#else
static Boolean
SetValuesPuzzleGL(Widget current, Widget request, Widget renew)
{
	RubikGLWidget c = (RubikGLWidget) current, w = (RubikGLWidget) renew;
	Boolean redraw = False;

	if (w->rubik.view != c->rubik.view) {
		ResizePuzzleGL(w);
		redraw = True;
	}
	if (w->rubik.sizex != c->rubik.sizex ||
			w->rubik.sizey != c->rubik.sizey ||
			w->rubik.sizez != c->rubik.sizez) {
		ResetPieces((RubikWidget) w);
		ResizePuzzleGL(w);
		redraw = True;
	}
#if 0
	if (w->rubikGL.cubeSizex.x != c->rubikGL.cubeSizex.x ||
			w->rubikGL.cubeSizey.x != c->rubikGL.cubeSizey.x ||
			w->rubikGL.cubeSizez.x != c->rubikGL.cubeSizez.x ||
			w->rubikGL.cubeSizex.y != c->rubikGL.cubeSizex.y ||
			w->rubikGL.cubeSizey.y != c->rubikGL.cubeSizey.y ||
			w->rubikGL.cubeSizez.y != c->rubikGL.cubeSizez.y) {
		ResizePuzzleGL(w);
		redraw = True;
	}
#endif
	if (w->rubik.delay != c->rubik.delay) {
		w->rubikGL.numTurnInc = ((w->rubik.delay < MAXTURNINC) ?
			w->rubik.delay + 1 : MAXTURNINC);
	}
	return (redraw);
}

static GLXContext *
initGL(RubikGLWidget w) {
	XVisualInfo xviIn, *xviOut;
	int numVis;

	/*XGetWindowAttributes(XtDisplay(w), XtWindow(w), &xgwa); */
	xviIn.screen = DefaultScreen(XtDisplay(w));
	xviIn.visualid = XVisualIDFromVisual(DefaultVisual(XtDisplay(w),
		xviIn.screen));
	xviOut = XGetVisualInfo(XtDisplay(w), VisualScreenMask | VisualIDMask,
		&xviIn, &numVis);
	if (!xviOut) {
		XtWarning("Could not get XVisualInfo");
		return (GLXContext *) NULL;
	}
	if (glXContext)
		free(glXContext);
	if ((glXContext = (GLXContext *) malloc(sizeof (GLXContext))) ==
			NULL) {
		DISPLAY_ERROR("Not enough memory for glx info, exiting.");
	}
	*glXContext = glXCreateContext(XtDisplay(w), xviOut, 0, GL_TRUE);
	(void) XFree((char *) xviOut);
	if (!*glXContext) {
		XtWarning("Could not create rendering context");
		return (GLXContext *) NULL;
	}
	return glXContext;
}
#endif

#ifndef WINVER
static
#endif
void
ResizePuzzleGL(RubikGLWidget w)
{
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	glViewport(0, 0, (GLint) w->core.width, (GLint) w->core.height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1.0, 1.0, -1.0, 1.0, 5.0, 15.0);
	glMatrixMode(GL_MODELVIEW);
}

static void
makeCurrentGL(RubikGLWidget w)
{
	GLboolean rgbaMode;

#ifdef WINVER
	wglMakeCurrent(w->core.hDC, hRC);
#else
	if (!glXMakeCurrent(XtDisplay(w), XtWindow(w), *glXContext)) {
		DISPLAY_WARNING("GL error");
	}
#endif
	madeCurrent = True;
	/* True Color junk */
#if 1
	glGetBooleanv(GL_RGBA_MODE, &rgbaMode);
	if (!rgbaMode)
#endif
	{
#ifdef WINVER
		glClearIndex(0.0);
#else
		glIndexi(WhitePixel(XtDisplay(w),
			DefaultScreen(XtDisplay(w))));
		glClearIndex((float) BlackPixel(XtDisplay(w),
			DefaultScreen(XtDisplay(w))));
#endif
	}
	ResizePuzzleGL(w);

	glDrawBuffer(GL_BACK);
	glClearDepth(1.0);
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glColor3f(1.0, 1.0, 1.0);

	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0, GL_POSITION, position0);
	glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT1, GL_POSITION, position1);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
	glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glEnable(GL_CULL_FACE);
	glShadeModel(GL_FLAT);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_shininess);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_specular);

	w->rubikGL.step = 90;
#if 0
	w->rubikGL.vx = 0.05;
	if (LRAND() & 1)
		w->rubikGL.vx *= -1.0;
	w->rubikGL.vy = 0.05;
	if (LRAND() & 1)
		w->rubikGL.vy *= -1.0;
#endif
	w->rubikGL.rotateStep = 0.0;
}

#ifndef WINVER
static void
InitializePuzzleGL(Widget request, Widget renew)
{
	RubikGLWidget w = (RubikGLWidget) renew;

	SetAllColors((RubikWidget) w);
	(void) initGL(w);
	w->rubik.dim = 4; /* 2 and 3 already taken */
	w->rubikGL.numTurnInc = ((w->rubik.delay < MAXTURNINC) ?
		w->rubik.delay + 1 : MAXTURNINC);
	ResizePuzzleGL(w);
}
#endif

#ifndef WINVER
static
#endif
void
ExposePuzzleGL(
#ifdef WINVER 
RubikGLWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifdef WINVER
	if (hRC == NULL) {
#if 1
		BYTE type = PFD_TYPE_RGBA;
#else
		BYTE type = PFD_TYPE_COLORINDEX;
#endif
		DWORD flags = PFD_DOUBLEBUFFER;
		(void) setupPixelFormat(w, type, flags);
		hRC = wglCreateContext(w->core.hDC);
		w->rubikGL.movement.face = NO_FACE;
		w->rubikGL.movement.position = NO_POSITION;
		w->rubikGL.movement.direction = NO_DEPTH;
		w->rubikGL.rotateStep = 0.0;
		w->rubikGL.angleStepX = 0.0;
		w->rubikGL.angleStepY = 0.0;
		w->rubikGL.angleStepZ = 0.0;
	}
#else
	RubikGLWidget w = (RubikGLWidget) renew;

	if (!w->core.visible)
		return;
#endif
	if (!madeCurrent) {
		makeCurrentGL(w);
	}
	DrawFrameGL(w, w->rubik.focus);
	DrawAllPiecesGL(w);
}

#ifndef WINVER
static void
MovePuzzleGLTl(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		TL,
		(int) (event->xkey.state & ControlMask));
}

static void
MovePuzzleGLTop(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		TOP,
		(int) (event->xkey.state & ControlMask));
}

static void
MovePuzzleGLTr(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		TR,
		(int) (event->xkey.state & ControlMask));
}

static void
MovePuzzleGLLeft(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		LEFT,
		(int) (event->xkey.state & ControlMask));
}

static void
MovePuzzleGLRight(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		RIGHT,
		(int) (event->xkey.state & ControlMask));
}

static void
MovePuzzleGLBl(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		BL,
		(int) (event->xkey.state & ControlMask));
}

static void
MovePuzzleGLBottom(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		BOTTOM,
		(int) (event->xkey.state & ControlMask));
}

static void
MovePuzzleGLBr(RubikGLWidget w, XEvent *event, char **args, int nArgs)
{
	MovePuzzleInput((RubikWidget) w, event->xbutton.x, event->xbutton.y,
		BR,
		(int) (event->xkey.state & ControlMask));
}
#endif
#endif
